import struct
from PIL import Image
import os
import glob
import sys
import traceback
import subprocess
import wave
import re

# fn - absolute filename
# an - absolute path with name (no extension)
# ad - absolute path to directory
# rn - relative path with name (no extension)
# rf - relative file name

def binary(short):
    st = ""
    for ibit in range(15,-1,-1):
        if (1 << ibit) & short == 0:
            st = st + "0"
        else:
            st = st + "1"
    return st

def EnsureAd(ad):
    if not os.path.exists(ad):
        os.mkdir(ad)


class Pic(object):
    "Holds a pil image and a filename, and can save the image."
    def __init__(self, image, ad, rn, xCenter, yCenter):
        self.image = image
        self.ad = ad
        self.rn = rn
        self.x = 0
        self.y = 0
        self.xCenter = xCenter
        self.yCenter = yCenter
        (self.w, self.h) = image.size

    def PicWithName(self, ad, rn):
        return Pic(self.image, ad, rn, self.xCenter, self.yCenter)
    
    def XLeft(self):
        return -self.xCenter
    def XRight(self):
        return self.w - self.xCenter
    def YTop(self):
        return -self.yCenter
    def YBottom(self):
        return self.h - self.yCenter

    def PicInRect(self, ad, rn, w, h, xCenter, yCenter, pal):
        if self.image.mode == "P":
            image = Image.new("P", (w, h), 0)
            image.putpalette(pal)
        else:
            image = Image.new("RGBA", (w, h))
        x = xCenter - self.xCenter
        y = yCenter - self.yCenter
        image.paste(self.image, (x, y))
        return Pic(image, ad, rn, xCenter, yCenter)
        
    def Save(self, fGif = False):
        print "Saving " + self.Rf(fGif)
        if self.image.mode == "P":
            self.image.save(self.Fn(fGif), transparency=0)
        else:
            if fGif:
                image = self.image.convert("RGB")
                image = image.convert("P", palette = Image.ADAPTIVE, colors=255)
                image.paste(255, (0,0), self.ImageMask(True))
                image.save(self.Fn(True), transparency=255)
            else:
                self.image.save(self.Fn(False))
        
    def PutPixel(self, col):
        "Put a pixel in the image, autoincrementing x and y."
        if not self.FMore():
            self.Save()
            raise Exception("putpixel past end of pic!")
            
        self.image.putpixel((self.x, self.y), col)
        self.x = self.x + 1
        if self.x == self.w:
            self.x = 0
            self.y = self.y + 1
    
    def FMore(self):
        "Returns true if there are any more pixels to put."
        if self.y == self.h:
            return False
        return True
    
    def ImageMask(self, fInvert = False):
        if self.image.mode == "P" or fInvert:
            imgMask = Image.new("1", (self.w, self.h))
            for x in range(self.w):
                for y in range(self.h):
                    if self.image.mode == "P":
                        fTransparent = self.image.getpixel((x, y)) == 0
                    else:
                        assert self.image.mode == "RGBA"
                        fTransparent = self.image.getpixel((x, y))[3] == 0
                    if fInvert: fTransparent = not fTransparent
                    imgMask.putpixel((x, y), 0 if fTransparent else 1)
            return imgMask
        return self.image

    def Rf(self, fGif = False):
        return self.rn + (".gif" if fGif else ".png")
    def Fn(self, fGif = False):
        return self.ad + "/" + self.Rf(fGif)
        
class Anim(object):
    def __init__(self, stName, speed, rgipic):
        self.stName = stName
        self.speed = speed
        self.rgipic = rgipic

    def FObj(self):
        return True

    def SaveGif(self, rgpic, ad, pal, adExe):
        if len(self.rgipic) > 1:
            adFrames = ad + "/" + RnValid(self.stName)
            EnsureAd(adFrames)
                
            xLeftMin = 0
            xRightMax = 0
            yTopMin = 0
            yBottomMax = 0
            
            for ipic in self.rgipic:
                pic = rgpic[ipic]
                if pic.XLeft() < xLeftMin:
                    xLeftMin = pic.XLeft()
                if pic.XRight() > xRightMax:
                    xRightMax = pic.XRight()
                if pic.YTop() < yTopMin:
                    yTopMin = pic.YTop()
                if pic.YBottom() > yBottomMax:
                    yBottomMax = pic.YBottom()
            xCenter = -xLeftMin
            yCenter = -yTopMin
            w = xCenter + xRightMax
            h = yCenter + yBottomMax
            for iipic, ipic in enumerate(self.rgipic):
                pic = rgpic[ipic].PicInRect(ad, RnValid(self.stName) + "/" + str(iipic), w, h, xCenter, yCenter, pal)
                pic.Save(True)  #gif
                pic.Save(False) #png
            rgarg = [adExe + "/gifsicle", "--delay", "10", "--disposal", "bg", "--loopcount"]
            rgarg.extend([adFrames + "/" + str(iipic) + ".gif" for iipic in range(len(self.rgipic))])
            rfGif = subprocess.Popen(rgarg, stdout = subprocess.PIPE).communicate()[0]
            with open(ad + "/" + RnValid(self.stName) + ".gif", "wb") as f:
                f.write(rfGif)
        else:
            rgpic[self.rgipic[0]].PicWithName(ad, RnValid(self.stName)).Save() #png

    def Rgipic(self):
        return self.rgipic
    

class IPic(object):
    def __init__(self, stName, ipic):
        self.stName = stName
        self.ipic = ipic
        
    def SaveGif(self, rgpic, ad, pal, adExe):
        rgpic[self.ipic].PicWithName(ad, RnValid(self.stName)).Save()

    def FObj(self):
        return False

    def Rgipic(self):
        yield self.ipic
        
class Pal(object):
    def __init__(self, fnPngPal):
        self.image = Image.open(fnPngPal)
    
    def Palette(self):
        "Generate a palette for pil (list of 768 integers) based on a colour grid."
        (w, h) = self.image.size
        wBox = w / 16
        hBox = h / 16
        palette = []
        # {x,y}p - "palette-based" xy value (0-15)
        for xp in range(16):
            for yp in range(16):
                palette.extend(self.image.getpixel((xp * wBox, yp * hBox)))
        return palette

class KnpExtractor(object):
    def __init__(self, an, pal, adExe, fRaw):
        self.pal = pal
        self.an = an
        self.adExe = adExe
        self.fRaw = fRaw

    def Fn(self, ext):
        return self.an + "." + ext
    
    def FnImg(self):
        return self.Fn("img")

    def FnGam(self):
        return self.Fn("gam")

    def FnSnd(self):
        return self.Fn("snd")

    def FnMus(self):
        return self.Fn("mus")
    
    def Rgfn(self):
        return glob.glob(self.an + ".*")
    
    def AdOut(self):
        return self.an
                            
    def Extract(self):
        EnsureAd(self.AdOut())
        if self.fRaw:
            EnsureAd(self.AdOut() + "/raw")
            for pic in self.Rgpic():
                if pic != None:
                    pic.Save()
        else:
            rgpic = [pic for pic in self.Rgpic(True)]
            print "IMG file loaded, reading GAM"
            for anim in self.Rganim():
                anim.SaveGif(rgpic, self.AdOut(), self.pal.Palette(), self.adExe)
            self.ExtractSound()

    def ExtractSound(self):
        if os.path.exists(self.FnSnd()):
            isnd = 0
            for rfSnd in self.RgrfFromHdr(self.FnSnd()):
                if rfSnd != None:
                    cbSound = struct.unpack_from("<I", rfSnd, 6)
                    stName = self.StFromRf(rfSnd, 10, 8)
                    (cchan, hz, _bps, _blockalign, cbit_samp) = struct.unpack_from("<HIIHH", rfSnd, 34)
                    rfData = rfSnd[48:]
                    rn = RnValid(str(isnd) + "-" + stName + ".wav")
                    print "Saving", rn
                    wav = wave.open(self.AdOut() + "/" + rn, "wb")
                    wav.setframerate(hz)
                    wav.setnchannels(cchan)
                    wav.setsampwidth(cbit_samp / 8)
                    wav.writeframes(rfData)
                    wav.close()
                isnd = isnd + 1
        if os.path.exists(self.FnMus()):
            imus = 0
            for rfMus in self.RgrfFromHdr(self.FnMus()):
                if rfMus != None:
                    cbMidi = struct.unpack_from("<I", rfMus, 6)
                    stName = self.StFromRf(rfMus, 10, 8)
                    rn = RnValid(str(imus) + "-" + stName + ".mid")
                    print "Saving", rn
                    with open(self.AdOut() + "/" + rn, "wb") as f:
                        f.write(rfMus[32:])
                imus = imus + 1
        
    def Rganim(self):
        with open(self.FnGam(), "rb") as f:
            rfFile = f.read()
        fo = 348
        cframe = struct.unpack_from("<H", rfFile, fo)[0]
        fo = fo + 10
        for iframe in range(cframe):
            for anim in self.RganimFromFrame(rfFile, struct.unpack_from("<I", rfFile,fo)[0], iframe):
                yield anim
            fo = fo + 4

    def StFromRf(self, rf, fo = 0, len = -1):
        if len < 0:
            st = rf[fo:]
        else:
            st = rf[fo:fo + len]
        ist0 = st.find("\x00")
        if ist0 >= 0:
            st = st[:ist0]
        return st
    
    def RganimFromFrame(self, rfFile, fo, iframe):
        kframe = struct.unpack_from("<B", rfFile, fo + 50)[0]
        if kframe == 1:
            fo = fo + 118
            cobj = struct.unpack_from("<H", rfFile, fo)[0]
            fo = fo + 6
            for iobj in range(cobj):
                stName = self.StFromRf(rfFile, fo, 40)
                kobj = struct.unpack_from("<B", rfFile, fo + 42)[0]
                foExtra = fo + 62    
                if kobj < 8:
                    cbyteExtra = struct.unpack_from("<H", rfFile, foExtra)[0]
                    if kobj == 2: # active object
                        foTables = foExtra + struct.unpack_from("<H", rfFile, foExtra + 6)[0]
                        for ikpanim in range(1,16):
                            rfoKpanim = struct.unpack_from("<h", rfFile, foTables + 2 + (ikpanim * 2))[0]
                            if rfoKpanim > 0:
                                for anim in self.RganimFromKpanim(rfFile, foTables + rfoKpanim, iframe, stName, ikpanim):
                                    yield anim
                    elif kobj == 1: # backdrop
                        ipic = struct.unpack_from("<H", rfFile, fo + 54)[0]
                        yield IPic("-".join([str(iframe), str(ipic), stName]), ipic)
                    fo = foExtra + cbyteExtra + 6
                    # skip thumbnail
                    fo = fo + struct.unpack_from("<H", rfFile, fo)[0] + 18
                else:
                    fo = foExtra
                

    def RganimFromKpanim(self, rfFile, foKpanim, iframe, stName, ikpanim):
        for idir in range(32):
            foRfoAnim = foKpanim + (idir * 2)
            rfoAnim = struct.unpack_from("<h", rfFile, foRfoAnim)[0]
            if rfoAnim > 0:
                foAnim = foKpanim + rfoAnim
                (spdMin, spdMax, cRepeat, iFrameLoop, cImg) = struct.unpack_from("<BBHHH", rfFile, foAnim)
                rgipic = struct.unpack_from("<" + ("H" * cImg), rfFile[foAnim + 8:foAnim + 8 + (2 * cImg)])
                yield Anim("-".join([str(iframe), stName, self.mpikpanim_stName[ikpanim], str(idir)]), spdMax, rgipic)
                    
    mpikpanim_stName = [None, "Stopped", "Walking", "Running", "Appearing", "Disappearing", "Bouncing", "Shooting", "Jumping",
                        "Falling", "Climbing", "CrouchDown", "StandUp", "UserAnim1", "UserAnim2", "UserAnim3", "UserAnim4"]

    def RgrfFromHdr(self, fn):
        if os.path.exists(fn):
            with open(fn, "rb") as f:
                rfFile = f.read()
            if len(rfFile) > 4:
                (cimgHdr,) = struct.unpack_from("<I", rfFile);
                
                foImgHdr = 4
                for iimgHdr in range(cimgHdr):
                    (foImg, szImg) = struct.unpack_from("<II", rfFile, foImgHdr)
                    if foImg > 0:
                        yield rfFile[foImg:foImg + szImg]
                    else:
                        yield None
                    foImgHdr = foImgHdr + 8
            
    def Rgpic(self, fDisplayProgress = False):
        if fDisplayProgress:            
            rgrfPic = [rfPic for rfPic in self.RgrfFromHdr(self.FnImg())]
            if len(rgrfPic) == 0: return
            stMeter = ("-" * 10).join(["0%", "25%", "50%", "75%", "100%"])
            print stMeter
            cdisp = 0.0
            ddisp = len(stMeter) / float(len(rgrfPic))
        else:
            rgrfPic = self.RgrfFromHdr(self.FnImg())

        for irfPic, rfPic in enumerate(rgrfPic):
            if rfPic != None:
                yield self.PicFromRf("raw/" + str(irfPic), rfPic, 0)
            else:
                yield None
            if fDisplayProgress:
                cdispT = cdisp + ddisp
                if int(cdispT) > int(cdisp):
                    getattr(__builtins__, "print")("." * (int(cdispT) - int(cdisp)), sep='', end='')
                cdisp = cdispT
            
        if fDisplayProgress:
            print

    class Info(object):
        def __init__(self, rfFile, fo = 0):
            (self.szPic, self.width, self.height) = struct.unpack_from("<IHH", rfFile, fo + 6)
            (self.cmpf, self.hotX, self.hotY) = struct.unpack_from("<Bhh", rfFile, fo + 15) # hotspots and action points can be negative(!)
            self.cbPix = 1 if self.cmpf != 2 else 2
            if self.cbPix == 1 and (self.width % 2) == 1:
                self.width = self.width + 1
            
    def PicFromRf(self, rn, rfFile, fo):
        info = self.Info(rfFile, fo)

        if info.cbPix == 1:
            image = Image.new("P", (info.width, info.height), 0)
            image.putpalette(self.pal.Palette())
        else:
            image = Image.new("RGBA", (info.width, info.height))
        pic = Pic(image, self.AdOut(), rn, info.hotX, info.hotY)


        rfImdat = None        
        if info.cmpf == 0:
            rfImdat = rfFile[fo + 24:fo + 24 + info.szPic]
        elif info.cmpf == 1 or info.cmpf == 2:
            rfImdat = self.RfImdatFromRle(pic, rfFile, fo + 24, info.szPic, info.cbPix)
        else:
            print "Invalid compression format", info.cmpf, "for image ", rn

        if rfImdat:
            self.PutImageRaw(pic, rfImdat, info.cbPix)
        
        return pic
        
    def PutImageRaw(self, pic, rfImdat, cbPix):
        if cbPix == 1:
            pic.image.putdata(rfImdat)
        else:
            fo = 0
            while fo < len(rfImdat):
                colr = struct.unpack_from("<H", rfImdat, fo)[0]
                if colr == 0:
                    pic.PutPixel((0,0,0,0))
                else:
                    pic.PutPixel(((colr & 0x7c00) >> 7, (colr & 0x03e0) >> 2, (colr & 0x1F) << 3, 255))
                if colr > 0x7fff:
                    print "full 16 bit colour used", colr
                fo = fo + 2
    
    def RfImdatFromRle(self, pic, rfFile, foImdat, sz, cbPix):
        try:
            rgrf = []
            foImdatLim = foImdat + sz
            while foImdat < foImdatLim:
                ctl = ord(rfFile[foImdat])
                if ctl == 0:
                    if (foImdat < foImdatLim - 1):
                        print "early end of image for ", pic.Rf()
                    break
                foImdat = foImdat + 1
                if (ctl & 0x80) == 0: # rle
                    rgrf.append(rfFile[foImdat:foImdat + cbPix] * ctl)
                    foImdat = foImdat + cbPix
                else:
                    cpixUnique = ctl & 0x7f
                    rgrf.append(rfFile[foImdat:foImdat + (cpixUnique * cbPix)])
                    foImdat = foImdat + (cpixUnique * cbPix)
            return ''.join(rgrf)
        except:
            print "error reading", pic.Rf()

class KnpLoader(object):
    def __init__(self, adExe, rgarg):
        self.pal = Pal(adExe + "/knp_pal.png")
        self.fRaw = False
        self.rgarg = []
        for arg in rgarg:
            if arg.lower() == "--raw":
                self.fRaw = True
            else:
                self.rgarg.append(arg)
        self.adExe = adExe
    
    def RgExtractor(self):
        if (len(self.rgarg) == 0):
            for extractor in self.ScanAd(os.getcwd()):
                yield extractor
        else:
            for arg in self.rgarg:
                fnImg = self.AnFromFn(arg) + ".img"
                if os.path.exists(fnImg):
                    yield self.Extractor(fnImg)
                elif os.path.isdir(arg):
                    for extractor in self.ScanAd(arg):
                        yield extractor
    
    
    def ScanAd(self, ad):
        for fnImg in glob.glob(ad + "/*.img"):
            yield self.Extractor(fnImg)

    def AnFromFn(self, fn):
        if fn.lower().endswith(".img") or fn.lower().endswith(".gam") or fn.lower().endswith(".exe"):
            return fn[:-4]
        return fn
    
    def Extractor(self, fn):
        return KnpExtractor(self.AnFromFn(fn), self.pal, self.adExe, self.fRaw)
        
def main(adExe, rgarg):
    print "KNPExtract v1.0"
    print "By Jeremy Penner"
    print "This code is released under the WTFPLv2, see license.txt for details"
    print
    print "Usage:"
    print "KNPExtract [--raw] [directory to scan for .gam files | .gam file] ..."

    ldr = KnpLoader(adExe, rgarg)
    for extractor in ldr.RgExtractor():
        print
        print "Extracting from " + os.path.basename(extractor.FnImg()) + "..."
        try:
            extractor.Extract()
        except Exception as e:
            print "Error:"
            traceback.print_exc()
            

def AdExe(fnExe):
    return os.path.dirname(os.path.abspath(sys.argv[0]))

def RnValid(rn):
    return re.sub('[^a-zA-Z0-9_\-.() ]', '_', rn)
if __name__ == "__main__":
    main(AdExe(sys.argv[0]), sys.argv[1:])