import knpextract
from knpextract import Pic, KnpLoader, AdExe, RnValid
from PIL import Image
import random
import os
import sys
import glob
from datetime import datetime

class Rig(object):
	def __init__(self, ad, rnd = random.Random()):
		self.ad = ad
		self.rgimd = [Imd(ad + "/gen_big", 1,2, 40000),
						Imd(ad + "/gen_med", 1,4, 2000, 40000),
						Imd(ad + "/gen_small", 2,7, 0, 2000)]
		self.rnd = rnd
	
	def PastePic(self, pic):
		x = self.rnd.randint(0, abs(self.pic.w - pic.w))
		y = self.rnd.randint(0, abs(self.pic.h - pic.h))
		self.pic.image.paste(pic.image, (x, y), pic.ImageMask())

	def PicGenerate(self):
		self.pic = Pic(Image.new("RGBA", (640,480)), self.ad, datetime.now().strftime("%d_%b_%Y-%H_%M_%f"), 0, 0)
		for imd in self.rgimd:
			for pic in imd.RgpicRand():
				self.PastePic(pic)
		return self.pic

	def SavePic(self, pic):
		for imd in self.rgimd:
			if imd.FInImd(pic):
				imd.SavePic(pic)

class Imd(object): # image directory
	def __init__(self, ad, cpicMin, cpicMax, sizeMin, sizeLim = -1):
		self.ad = ad
		self.sizeMin = sizeMin
		self.sizeLim = sizeLim
		self.cpicMin = cpicMin
		self.cpicMax = cpicMax
		if not os.path.isdir(ad):
			os.mkdir(ad)

	def FInImd(self, pic):
		size = pic.w * pic.h
		return size >= self.sizeMin and (self.sizeLim < 0 or (size < self.sizeLim))
	
	def SavePic(self, pic):
		rnPrefix = os.path.split(pic.ad)[1]
		picNew = pic.PicWithName(self.ad, rnPrefix + "_" + RnValid(pic.rn))
		picNew.Save()
		
	def PicRand(self):
		rgfnPic = glob.glob(self.ad + "/*.png")
		if len(rgfnPic) == 0:
			raise Exception("Please extract some more images!")
		fnPic = rgfnPic[random.randint(0, len(rgfnPic) - 1)]
		image = Image.open(fnPic)
		return Pic(image, self.ad, os.path.split(fnPic)[1], 0, 0)
	
	def RgpicRand(self):
		for ipic in range(random.randint(self.cpicMin, self.cpicMax)):
			yield self.PicRand()

def main(adExe, rgarg):
	print "KNPGen v1.0"
	print "By Jeremy Penner"
	print "This code is released under the WTFPLv2, see license.txt for details"
	print
	print "Usage:"
	print "(To extract images for use by KNPGen -- you only need to do this once for each IMG library)"
	print "KNPGen [directory to scan for .img files | .img file] ..."
	print
	print "(To generate a new image after images have been extracted)"
	print "KNPGen"
	print
	
	fExtract = len(rgarg) > 0 or not os.path.isdir(adExe + "/gen_small")
	rig = Rig(adExe)
	if (fExtract):
		ldr = KnpLoader(adExe, rgarg)
		rgpic = []
		for extractor in ldr.RgExtractor():
			print "Extracting from " + os.path.basename(extractor.FnImg()) + "..."
			for pic in extractor.Rgpic():
				rig.SavePic(pic)
	rig.PicGenerate().Save()

if __name__ == "__main__":
	main(AdExe(sys.argv[0]), sys.argv[1:])