from knpextract import *
import shutil
import random

# font manglers:
# "wingdingify"
# scramble

# sound manglers
# scramble
# glitch
# reverse

# music manglers
# scramble

class KNPMangler(object):
    def __init__(self, ad, knpextractor):
        self.adMangle = ad
        self.knpextractor = knpextractor
        self.BuildMaps()

    def MangleGame(self, mangconf):
        ad = self.adMangle
        imangle = 1
        while os.path.exists(ad):
            imangle = imangle + 1
            ad = self.adMangle + " " + str(imangle)
        os.mkdir(ad)
        for ext in ["gam", "exe"]:
            fn = self.knpextractor.Fn(ext)
            if os.path.exists(fn):
                shutil.copy(fn, ad + "/" + os.path.basename(fn))

        for ext in mpext_stRes.keys():
            rgrf = self.RgrfMangled(mangconf.rgmangler, ext)
            if rgrf != None:
                self.SaveRes(ad + "/" + os.path.basename(self.knpextractor.Fn(ext)), rgrf)

    def RgrfMangled(self, rgmangler, ext):
        fnOrig = self.knpextractor.Fn(ext)
        if os.path.exists(fnOrig):
            rgrf = list(self.knpextractor.RgrfFromHdr(fnOrig))
            for mangler in rgmangler:
                if mangler.ext == ext:
                    rgrf = mangler.Mangle(rgrf, self, self.knpextractor)
            return rgrf
        return None
    
    def RgvalHdr(self, rgrfImg):
        yield len(rgrfImg)
        fo = 4 + (8 * len(rgrfImg))
        for rfImg in rgrfImg:
            if rfImg == None:
                yield 0
                yield 0
            else:
                yield fo
                yield len(rfImg)
                fo = fo + len(rfImg)
            
    def SaveRes(self, fn, rgrf):
        with open(fn, "wb") as f:
            fo = 4 + (8 * len(rgrf))
            f.write(struct.pack("<I" + ("II" * len(rgrf)), *[val for val in self.RgvalHdr(rgrf)]))
            for rf in rgrf:
                if rf != None:
                    f.write(rf)

    def ColfFromPal(self, ipal, pal):
        return (pal[ipal * 3], pal[(ipal * 3) + 1], pal[(ipal * 3) + 2])
    def DColf(self, colf, colf2):
        return abs(colf[0] - colf2[0]) + abs(colf[1] - colf2[1]) + abs(colf[2] - colf2[2])
    
    def BuildMaps(self):
        pal = self.knpextractor.pal.Palette()
        rgcolf = []
        for ipal in range(1,256):
            rgcolf.append(self.ColfFromPal(ipal, pal))

        self.mpcolcol_inverse = [0]
        for icolf, colf in enumerate(rgcolf):
            colfInverse = (255 - colf[0], 255 - colf[1], 255 - colf[2])
            icolfInv = icolf
            dcolfMin = self.DColf(colf, colfInverse)
            for icolfT, colfT in enumerate(rgcolf):
                dcolf = self.DColf(colfInverse, colfT)
                if dcolf < dcolfMin:
                    icolfInv = icolfT
                    dcolfMin = dcolf
            self.mpcolcol_inverse.append(icolfInv + 1)

        self.mpcolcol_silhouette = [0]
        self.mpcolcol_silhouette.extend([160] * 255)
        
        self.mpcolcol_hole = [32]
        self.mpcolcol_hole.extend([0] * 255)

    def mapcol(self, rfImg, fo, mpcolcol):
        return chr(mpcolcol[ord(rfImg[fo])])

    class Rfex(object):
        def __init__(self, rf):
            self.rf = rf
            self.fo = 0
            self.foPrev = 0
            self.rgrf = []
        def SeekTo(self, fo):
            self.fo = fo
        def ReadByte(self, cb = 1):
            self.fo = self.fo + cb
            return ord(self.rf[self.fo - 1])
        def ReplaceByte(self, b):
            if self.foPrev < self.fo - 1:
                self.rgrf.append(self.rf[self.foPrev:self.fo - 1])
            self.rgrf.append(chr(b))
            self.foPrev = self.fo
        def ReplaceNextByte(self, mpb_bNew):
            self.ReplaceByte(mpb_bNew[self.ReadByte()])
        def MarkTransition(self):
            if self.foPrev != self.fo:
                self.rgrf.append(self.rf[self.foPrev:self.fo])
                self.foPrev = self.fo
        def RfNew(self):
            self.rgrf.append(self.rf[self.foPrev:])
            return ''.join(self.rgrf)

    def MangleColours(self, rfImg, mpcolcol):
        info = self.knpextractor.Info(rfImg)
        rfex = self.Rfex(rfImg)
        rfex.SeekTo(24)
        while rfex.fo < 24 + info.szPic:
            if info.cmpf == 0:
                rfex.ReplaceNextByte(mpcolcol)
            elif info.cmpf == 1:
                ctl = rfex.ReadByte()
                if ctl == 0:
                    break
                if (ctl & 0x80) == 0: # rle
                    rfex.ReplaceNextByte(mpcolcol)
                else:
                    cpixUnique = ctl & 0x7f
                    for ipixUnique in range(cpixUnique):
                        rfex.ReplaceNextByte(mpcolcol)
            else:
                raise "Unknown compression type:", info.cmpf
        return rfex.RfNew()
    
    def GlitchImage(self, rfImg, cglitch):
        info = self.knpextractor.Info(rfImg)
        rfex = self.Rfex(rfImg)
        rfex.SeekTo(24)
        rfex.MarkTransition()
        while rfex.fo < 24 + info.szPic:
            if info.cmpf == 0:
                rfex.ReadByte()
                if random.randrange(100) < 10:
                    rfex.MarkTransition()
            elif info.cmpf == 1 or info.cmpf == 2:
                cbPix = info.cmpf
                if random.randrange(100) < 75:
                    rfex.MarkTransition()
                ctl = rfex.ReadByte()
                if ctl == 0:
                    rfex.ReplaceByte(0) #mark transition before final 0
                    break
                if (ctl & 0x80) == 0: # rle
                    rfex.ReadByte(cbPix)
                else:
                    cpixUnique = ctl & 0x7f
                    for ipixUnique in range(cpixUnique):
                        rfex.ReadByte(cbPix)
            else:
                raise "Unknown compression type:", info.cmpf
        rfex.MarkTransition()
        if len(rfex.rgrf) > 3:
            for iglitch in range(cglitch):
                irf1 = random.randrange(1,len(rfex.rgrf) - 1)
                irf2 = random.randrange(1,len(rfex.rgrf) - 1)
                rf = rfex.rgrf[irf1]
                del rfex.rgrf[irf1]
                rfex.rgrf.insert(irf2, rf)
        return rfex.RfNew()


class Mangler(object):
    def __init__(self, stCmd, stDesc, ext, dgMangle):
        self.stCmd = stCmd
        self.stDesc = stDesc
        self.ext = ext
        self.dgMangle = dgMangle
        self.rgfilter = []

    def AddFilter(self, filter, *rgarg):
        self.rgfilter.append((filter, rgarg))    

    def FInclude(self, ipic, extractor):
        for filter in self.rgfilter:
            if filter[0](ipic, extractor, *(filter[1])):
                return False
        return True
    
    def Mangle(self, rgrfPic, mangler, extractor):
        rgpicFiltered = []
        rgrfPicT = []
        for ipic, rfPic in enumerate(rgrfPic):
            if rfPic == None or not self.FInclude(ipic, extractor):
                rgpicFiltered.append((ipic, rfPic))
            else:
                rgrfPicT.append(rfPic)
                
        rgrfPicNew = list(self.dgMangle(rgrfPicT, mangler, extractor))
        for ipic, rfPic in rgpicFiltered:
            rgrfPicNew.insert(ipic, rfPic)
        return rgrfPicNew

def FilterPct(ipic, extractor, pct):
    return random.randrange(100) >= pct

class Filto(object):
    def __init__(self):
        self.setIpicObj = None
        self.extractor = None
    def FilterObj(self, ipic, extractor):
        self.EnsureSet(extractor)
        return not (ipic in self.setIpicObj)
    def FilterNonObj(self, ipic, extractor):
        self.EnsureSet(extractor)
        return ipic in self.setIpicObj
    def EnsureSet(self, extractor):
        if self.setIpicObj == None or self.extractor != extractor:
            self.setIpicObj = set()
            self.extractor = extractor
            for anim in extractor.Rganim():
                if anim.FObj():
                    self.setIpicObj = self.setIpicObj.union(anim.Rgipic())

filto = Filto()

rgextRes = ["img", "mus", "snd", "mtf"]
rgextGame = ["gam", "exe"]
rgrnDll = ["knpg", "knps"]
mpext_stRes = {"img": "images", "snd": "sounds", "mus": "music", "mtf": "fonts"}

rgmanglerI = []
def Rgmangler():
    return rgmanglerI
            
def Reg(stCmd, stDesc, *rgext):
    def RegI(dg):
        for ext in rgext:
            if len(rgext) > 1:
                stCmdT = stCmd + ext
                stDescT = stDesc + mpext_stRes[ext]
            else:
                stCmdT = stCmd
                stDescT = stDesc
            mangler = Mangler(stCmdT, stDescT, ext, dg)
            rgmanglerI.append(mangler)
    return RegI

@Reg("scramble", "randomize order of ", *rgextRes)
def MangleScramble(rgrfPic, mangler, extractor):
    random.shuffle(rgrfPic)
    return rgrfPic

@Reg("silhouette", "turns images dark gray", "img")
def MangleSilhouette(rgrfPic, mangler, extractor):
    for rfPic in rgrfPic:
        yield mangler.MangleColours(rfPic, mangler.mpcolcol_silhouette)

@Reg("inverse", "make images inverse colours", "img")
def MangleInverse(rgrfPic, mangler, extractor):
    for rfPic in rgrfPic:
        yield mangler.MangleColours(rfPic, mangler.mpcolcol_inverse)

@Reg("glitch", "move chunks of images around randomly", "img")
def MangleGlitch(rgrfPic, mangler, extractor):
    for rfPic in rgrfPic:
        yield mangler.GlitchImage(rfPic, random.randrange(15))

# -- font format not yet understood -- this doesn't work
#@Reg("wingding", "turn fonts into wingdings", "mtf")
def MangleWingding(rgrfFont, mangler, extractor):
    for rfFont in rgrfFont:
        yield rfFont[:32] + "Wingdings\x00" + rfFont[42:]

#fromother usecases:
# --fromother --pct=20 foo --pct=30 --scrambleimg --glitch bar --pct=20 --objonly
# filters can apply to:
#  - the fromother mangler itself
#  - the games that fromother processes
#  - the manglers specified for each game

class Mangconf(object):
    GAME = 0
    MANGLER = 1
    FILTER = 2
    FROMOTHER = 3
    def __init__(self, adExe, rgarg):
        self.rgmangler = []
        self.rgargExtractor = []
        self.rgfromother = []
        self.adExe = adExe
        for arg in rgarg:
            self.AddArg(*self.ParseArg(arg))

    def FNoop(self):
        return len(self.rgmangler) == 0 and all(map(lambda fo: fo.FNoop(), self.rgfromother))

    def ParseArg(self, arg):
        if arg.startswith("--"):
            for mangler in Rgmangler():
                if arg.lower().startswith("--" + mangler.stCmd):
                    return (Mangconf.MANGLER, mangler)
            if arg.lower().startswith("--fromother"):
                return (Mangconf.FROMOTHER, arg[len("--fromother"):])
            elif arg.lower().startswith("--pct="):
                return (Mangconf.FILTER, (FilterPct, int(arg[6:])))
            elif arg.lower().startswith("--objonly"):
                return (Mangconf.FILTER, (filto.FilterObj,))
            elif arg.lower().startswith("--nonobj"):
                return (Mangconf.FILTER, (filto.FilterNonObj,))
        return (Mangconf.GAME, arg)

    def AddArg(self, kconf, val):
        if kconf == Mangconf.FROMOTHER:
            if val.lower() in rgextRes:            
                fromother = Fromother(val, self.adExe)
                self.rgfromother.append(fromother)
                self.rgmangler.append(fromother.manglerFO)
            else:
                print "ignored fromother --", val, "is not a valid resource type"
        elif len(self.rgfromother) > 0:
            self.rgfromother[-1].AddArg(kconf, val)
        elif kconf == Mangconf.GAME:
            self.rgargExtractor.append(val)
        elif kconf == Mangconf.MANGLER:
            self.rgmangler.append(val)
        elif kconf == Mangconf.FILTER:
            if len(self.rgmangler) > 0:
                self.rgmangler[-1].AddFilter(val[0], *val[1:])
            else:
                print "ignored filter", val[0], "-- nothing to apply it to"
    def Rgextractor(self):
        return KnpLoader(self.adExe, self.rgargExtractor).RgExtractor()
                
class Fromother(object):
    def __init__(self, ext, adExe):
        self.manglerFO = Mangler("fromother" + ext, "steals resources from another game", ext, self.Mangle)
        self.rgmangler = []
        self.rgmangconf = []
        self.adExe = adExe
        self.fLastGame = False

    def FNoop(self):
        return len(self.rgmangconf) == 0
    
    def AddArg(self, kconf, val):
        if kconf == Mangconf.GAME:
            if not self.fLastGame:
                self.rgmangconf.append(Mangconf(self.adExe, []))
            self.rgmangconf[-1].AddArg(kconf, val)
            self.fLastGame = True
        else:
            self.fLastGame = False
            if len(self.rgmangconf) > 0:
                self.rgmangconf[-1].AddArg(kconf, val)
            elif kconf == Mangconf.MANGLER:
                self.rgmangler.append(val)
            elif kconf == Mangconf.FILTER:
                if len(self.rgmangler) > 0:
                    self.rgmangler[-1].AddFilter(val[0], *val[1:])
                else:
                    self.manglerFO.AddFilter(val[0], *val[1:])

    def Mangle(self, rgrf, mangler, extractor):
        rgrfNew = []
        for mangconf in self.rgmangconf:
            for extractorT in mangconf.Rgextractor():
                for rfMangled in KNPMangler(mangler.adMangle, extractorT).RgrfMangled(mangconf.rgmangler, self.manglerFO.ext):
                    if rfMangled != None:
                        rgrfNew.append(rfMangled)
        if len(rgrfNew) == 0:
            return rgrf
        for manglerT in self.rgmangler:
            if manglerT.ext == self.manglerFO.ext:
                rgrfNew = list(manglerT.Mangle(rgrfNew, mangler, extractor))
        if len(rgrfNew) < len(rgrf):
            rgrfNew = rgrfNew * ((len(rgrf) / len(rgrfNew)) + 1)
        return rgrfNew[:len(rgrf)]
        
def AddFilters(rgmangler, filter, *rgarg):
    for mangler in rgmangler:
        mangler.AddFilter(filter, *rgarg)

def GenFO(extractorOrig, extractor, cmangler):
    while True:
        ext = rgextRes[random.randrange(len(rgextRes))]
        if os.path.exists(extractor.Fn(ext)) and os.path.exists(extractorOrig.Fn(ext)):
            break
    yield "--fromother" + ext
    yield "--pct=" + str(random.randrange(20,70))
    yield extractor.Fn(ext)
    if extractor == extractorOrig and cmangler == 0:
        cmangler = 1
    for imangler in range(cmangler):
        while True:
            mangler = Rgmangler()[random.randrange(len(Rgmangler()))]
            if mangler.ext == ext:
                yield "--" + mangler.stCmd
                break
    
def main(adExe, rgarg):
    print "KNPMangle v1.0"
    print "By Jeremy Penner"
    print "This code is released under the WTFPLv2, see license.txt for details"
    print
    print "Usage:"
    print "KNPMangle [directory to scan for .img files | .img file] ... (manglers) ..."

    mangconf = Mangconf(adExe, rgarg)

    if mangconf.FNoop():
        print
        print "Available manglers:"
        for mangler in Rgmangler():
            print str.format("--{0:20}{1}", mangler.stCmd, mangler.stDesc)
        print
        print "Available filters:"
        print "--objonly        only run mangler on images used by Active Objects"
        print "--nonobj         only run mangler on images not used by Active Objects"
        print "--pct=(1-100)    only run filter on a random portion of resources"
        print
        print "Filters apply to the last specified mangler.  For example, this commandline:"
        print "KNPMangle --silhouette --objonly --scrambleimg --pct=30 --objonly"
        print "would turn all objects into gray silhouettes, and then mix up the order of"
        print "30% of the object images."
        print
        print "You can use resources from other games with the --fromother[ext] command."
        print "See readme for details."
        print
        rgextractor = list(mangconf.Rgextractor())
        if len(rgextractor) == 0:
            print "No games found."
            return
        extractorOrig = rgextractor[random.randrange(len(rgextractor))]
        rgargGen = [extractorOrig.FnGam()]
        for ifo in range(random.randrange(1,4)):
            rgargGen.extend(GenFO(extractorOrig, rgextractor[random.randrange(len(rgextractor))], random.randrange(3)))
        print "Using commands:"
        print " ".join(rgargGen)
        print
        mangconf = Mangconf(adExe, rgargGen)
    
    for extractor in mangconf.Rgextractor():
        print
        print "Mangling " + os.path.basename(extractor.Fn("gam")) + "..."
        try:
            mangler = KNPMangler(extractor.AdOut(), extractor)
            mangler.MangleGame(mangconf)
        except Exception as e:
            print "Error:"
            traceback.print_exc()
            
if __name__ == "__main__":
    main(AdExe(sys.argv[0]), sys.argv[1:])